#include "ehs/io/socket/Request.h"
#include "ehs/Base64.h"

namespace ehs
{
	Request::Request()
		: verb(Verb::POST), cType(ContentType::NONE)
	{
	}

	Request::Request(const Verb &verb, Str_8 rsrc)
		: verb(verb), rsrc((Str_8 &&)rsrc), cType(ContentType::NONE)
	{
	}
	
	Request::Request(const char *data, const UInt_64 &size)
		: verb(Verb::POST), cType(ContentType::NONE)
	{
		ReadData(Str_8(data, size));
	}

	Request::Request(const Str_8 &data)
		: verb(Verb::POST), cType(ContentType::NONE)
	{
		ReadData(data);
	}

	Request::Request(Request&& other) noexcept
		: verb(other.verb), rsrc((Str_8 &&)other.rsrc), queries((Vector<QueryVar> &&)other.queries),
		header((Vector<HeaderVar> &&)other.header), cType(other.cType), body((Str_8 &&)other.body)
	{
		other.verb = Verb::POST;
		other.cType = ContentType::NONE;
	}

	Request::Request(const Request& other)
		: verb(other.verb), rsrc(other.rsrc), queries(other.queries), header(other.header), cType(other.cType),
		body(other.body)
	{
	}

	Request& Request::operator=(Request&& other) noexcept
	{
		if (this == &other)
			return *this;

		verb = other.verb;
		rsrc = (Str_8 &&)other.rsrc;
		queries = (Vector<QueryVar> &&)other.queries;
		header = (Vector<HeaderVar> &&)other.header;
		cType = other.cType;
		body = (Str_8 &&)other.body;

		other.verb = Verb::POST;
		other.cType = ContentType::NONE;

		return* this;
	}

	Request& Request::operator=(const Request &other)
	{
        if (this == &other)
            return *this;

		verb = other.verb;
		rsrc = other.rsrc;
		queries = other.queries;
		header = other.header;
        cType = other.cType;
		body = other.body;

		return* this;
	}

	Verb Request::GetVerb() const
	{
		return verb;
	}
	
	void Request::SetContentType(const ContentType &cType)
	{
		if (body.Size())
			body.Resize(0);
		
		this->cType = cType;
	}

	ContentType Request::GetContentType() const
	{
		return cType;
	}

	void Request::SetResource(Str_8 rsrc)
	{
		this->rsrc = (Str_8 &&)rsrc;
	}

	Str_8 Request::GetResource() const
	{
		return rsrc;
	}

	bool Request::HasQueryVar(const UInt_64& id) const
	{
		for (UInt_64 i = 0; i < queries.Size(); ++i)
			if (queries[i] == id)
				return true;

		return false;
	}

	bool Request::HasQueryVar(const Str_8& name) const
	{
		return HasQueryVar(name.Hash_64());
	}

	bool Request::AddQueryVar(QueryVar var)
	{
		if (HasQueryVar(var.GetId()))
			return false;

		queries.Push((QueryVar &&)var);

		return true;
	}

	bool Request::RemoveQueryVar(const UInt_64& id)
	{
		for (UInt_64 i = 0; i < queries.Size(); ++i)
		{
			if (queries[i] != id)
				continue;

			queries.Swap(i, queries.End());
			queries.Pop();

			return true;
		}

		return false;
	}

	bool Request::RemoveQueryVar(const Str_8& name)
	{
		return RemoveQueryVar(name.Hash_64());
	}

	QueryVar *Request::GetQueryVar(const UInt_64& id) const
	{
		for (UInt_64 i = 0; i < queries.Size(); ++i)
			if (queries[i] == id)
				return &queries[i];

		return nullptr;
	}

	QueryVar *Request::GetQueryVar(const Str_8& name) const
	{
		return GetQueryVar(name.Hash_64());
	}

	Str_8 Request::GetQueryValue(const UInt_64& id) const
	{
		for (UInt_64 i = 0; i < queries.Size(); ++i)
			if (queries[i] == id)
				return queries[i].GetValue();

		return {};
	}

	Str_8 Request::GetQueryValue(const Str_8& name) const
	{
		return GetQueryValue(name.Hash_64());
	}

	Vector<QueryVar> Request::GetQueries() const
	{
		return queries;
	}
	
	void Request::BasicAuth(const Str_8& id, const Str_8& secret)
	{
		AddHeaderVar({"Authorization", Str_8("Basic ") + Base64::Encode(id + ":" + secret)});
	}
	
	void Request::BearerAuth(const Str_8& token)
	{
		AddHeaderVar({"Authorization", "Bearer " + token});
	}

    void Request::BearerAuth(const Str_8& token, const Str_8& clientId)
    {
        AddHeaderVar({"Authorization", "Bearer " + token});
        AddHeaderVar({"Client-Id", clientId});
    }
	
	void Request::BotAuth(const Str_8& token)
	{
		AddHeaderVar({"Authorization", "Bot " + token});
	}

	bool Request::HasHeaderVar(const UInt_64& id) const
	{
		for (UInt_64 i = 0; i < header.Size(); ++i)
			if (header[i] == id)
				return true;

		return false;
	}

	bool Request::HasHeaderVar(const Str_8& name) const
	{
		return HasHeaderVar(name.Hash_64());
	}

	bool Request::AddHeaderVar(HeaderVar var)
	{
		if (HasHeaderVar(var.GetId()))
			return false;

		header.Push((HeaderVar &&)var);

		return true;
	}

	bool Request::RemoveHeaderVar(const UInt_64 &id)
	{
		for (UInt_64 i = 0; i < header.Size(); ++i)
		{
			if (header[i] != id)
				continue;

			header.Swap(i, header.End());
			header.Pop();

			return true;
		}

		return false;
	}

	bool Request::RemoveHeaderVar(const Str_8 &name)
	{
		return RemoveHeaderVar(name.Hash_64());
	}

	HeaderVar *Request::GetHeaderVar(const UInt_64 &id) const
	{
		for (UInt_64 i = 0; i < header.Size(); ++i)
			if (header[i] == id)
				return &header[i];

		return nullptr;
	}

	HeaderVar *Request::GetHeaderVar(const Str_8 &name) const
	{
		return GetHeaderVar(name.Hash_64());
	}

	Str_8 Request::GetHeaderValue(const UInt_64 &id) const
	{
		for (UInt_64 i = 0; i < header.Size(); ++i)
			if (header[i] == id)
				return header[i].GetValue();

		return {};
	}

	Str_8 Request::GetHeaderValue(const Str_8 &name) const
	{
		return GetHeaderValue(name.Hash_64());
	}

	Vector<HeaderVar> Request::GetHeader() const
	{
		return header;
	}
	
	void Request::AddToBody(const Str_8& var, const Str_8& value)
	{
		if (body.Size())
		{
			if (cType == ContentType::APP_FORMURLENCODED)
				body.Push('&');
		}
		
		body += var;
		
		if (cType == ContentType::APP_FORMURLENCODED)
			body += "=";
			
		body += value;
	}
	
	void Request::AddToBody(const Str_8& data)
	{
		if (body.Size())
		{
			if (cType == ContentType::APP_FORMURLENCODED)
				body.Push('&');
		}
		
		body += data;
	}
	
	void Request::SetBody(const Str_8& body)
	{
		this->body = body;
	}

	Str_8 Request::GetVar(const Str_8& var) const
	{
		Vector<Str_8> vars;

		if (cType == ContentType::APP_FORMURLENCODED)
			vars = body.Split("&");

		for (UInt_64 i = 0; i < vars.Size(); ++i)
		{
			Vector<Str_8> data = vars[i].Split("=");

			if (data[0] == var)
				return data[1];
		}

		return "";
	}
	
	Str_8 Request::GetBody() const
	{
		return body;
	}

	Json Request::GetJson() const
	{
		return {body, 5};
	}
	
	Str_8 Request::FormResult() const
	{
		Str_8 result = VerbToStr(verb) + " " + rsrc;

		if (queries.Size())
		    result += "?" + queries[0].ToStr();

		for (UInt_64 i = 1; i < queries.Size(); ++i)
			result += "&" + queries[i].ToStr();

		result += " HTTP/1.1\r\n";

		for (UInt_64 i = 0; i < header.Size(); ++i)
		{
			if (header[i] == 11245195881701120957ull || header[i] == 6556302699635904946ull ||
				header[i] == 13347276190653935221ull) // Content-Length, Server, and Content-Type in order from left to right.
				continue;

			result += header[i].ToStr() + "\r\n";
		}

		result += "Content-Type: " + ContentTypeToStr(cType) + "\r\n";

		if (verb == Verb::GET)
			result += "\r\n";
		else
			result += "Content-Length: " + Str_8::FromNum(body.Size()) + "\r\n\r\n" + body;

		return result;
	}

	bool Request::IsValid() const
	{
		return rsrc.Size() || queries.Size() || header.Size() || body.Size();
	}
	
	Str_8 Request::VerbToStr(const Verb &verb)
	{
		switch (verb)
		{
        case Verb::POST:
			return "POST";
		case Verb::GET:
			return "GET";
		case Verb::PUT:
			return "PUT";
		case Verb::DEL:
			return "DELETE";
		default:
			return "";
		}
	}
	
	Str_8 Request::ContentTypeToStr(const ContentType &cType)
	{
		switch (cType)
		{
		case ContentType::APP_MULTIPART_FORMDATA:
			return "multipart/form-data";
		case ContentType::APP_FORMURLENCODED:
			return "application/x-www-form-urlencoded";
		case ContentType::APP_JAVASCRIPT:
			return "application/javascript";
		case ContentType::APP_JSON:
			return "application/json";
		case ContentType::APP_XML:
			return "application/xml";
		case ContentType::TEXT_PLAIN:
			return "text/plain";
		case ContentType::TEXT_HTML:
			return "text/html";
		case ContentType::TEXT_CSS:
			return "text/css";
		case ContentType::TEXT_XML:
			return "text/xml";
		case ContentType::IMG_X_ICON:
			return "image/x-icon";
		default:
			return "";
		}
	}

	ContentType Request::StrToContentType(const Str_8 &value)
	{
		if (value == "multipart/form-data")
			return ContentType::APP_MULTIPART_FORMDATA;
		else if (value == "application/x-www-form-urlencoded")
			return ContentType::APP_FORMURLENCODED;
		else if (value == "application/javascript")
			return ContentType::APP_JAVASCRIPT;
		else if (value == "application/json")
			return ContentType::APP_JSON;
		else if (value == "application/xml")
			return ContentType::APP_XML;
		else if (value == "text/plain")
			return ContentType::TEXT_PLAIN;
		else if (value == "text/html")
			return ContentType::TEXT_HTML;
		else if (value == "text/css")
			return ContentType::TEXT_CSS;
		else if (value == "text/xml")
			return ContentType::TEXT_XML;
		else if (value == "image/x-icon")
			return ContentType::IMG_X_ICON;
		else
			return ContentType::NONE;
	}

	void Request::ReadData(const Str_8 &data)
	{
		Vector<Str_8> lines = data.Split("\r\n");
		Vector<Str_8> meta = lines[0].Split(" ");

		if (meta[0] == "POST")
			verb = Verb::POST;
		else if (meta[0] == "GET")
			verb = Verb::GET;
		else if (meta[0] == "PUT")
			verb = Verb::PUT;

		UInt_64 queryIndex = 0;
		if (meta[1].Find("?", &queryIndex))
		{
			rsrc = meta[1].Sub(0, queryIndex);
			cType = ContentType::APP_FORMURLENCODED;

			Vector<Str_8> strQueries = meta[1].Sub(queryIndex + 1).Split("&");
			for (UInt_64 i = 0; i < strQueries.Size(); ++i)
			{
				const Vector<Str_8> queryVar = strQueries[i].Split("=");

				queries.Push({queryVar[0], queryVar[1]});
			}
		}
		else
		{
			rsrc = meta[1];
		}

		for (UInt_64 i = 1; i < lines.Size(); ++i)
		{
			if (!lines[i].Size())
				break;

			Vector<Str_8> var = lines[i].Split(": ");

			if (var[0] == "Content-Type")
			{
				cType = StrToContentType(var[1]);
				continue;
			}

			if (var[0] == "Content-Length")
				continue;

			Vector<Str_8> headerVar = lines[i].Split(": ");

			header.Push({headerVar[0], headerVar[1]});
		}
	}
}